/* Arduino Si4703 Library
 * Written by Ryan Owens for SparkFun Electronics
 * 7/1/11
 *
 * This library is for use with the SparkFun Si4703 Shield
 * Released under the 'Buy Me a Beer' license
 * (If we ever meet, you buy me a beer)
 *
 * See the header file for better function documentation.
 *
 * See the example sketches to learn how to use the library in your code.
*/

#include "Si4703.h"
#include "si4703RegisterMap.h"
#include "WProgram.h"
#include "Wire.h"

/****************************************************************
*
* Public Functions
*
****************************************************************/

//This is just a constructor.
Si4703::Si4703(){
	//Initialize the local register values to 0
	for(int i=0; i<16; i++)si4703_registers[i]=0;
}


void Si4703::begin(void){

	//Configure the Si4703 for 2-wire communcation
	pinMode(RADIO_RESET_PIN, OUTPUT);
	pinMode(SDIO, OUTPUT); //SDIO is connected to A4 for I2C
	digitalWrite(SDIO, LOW); //A low SDIO indicates a 2-wire interface
	digitalWrite(RADIO_RESET_PIN, LOW); //Put Si4703 into reset
	delay(1); //Some delays while we allow pins to settle
	digitalWrite(RADIO_RESET_PIN, HIGH); //Bring Si4703 out of reset with SDIO set to low and SEN pulled high with on-board resistor
	delay(1); //Allow Si4703 to come out of reset

	//Configure the I2C connection
	Wire.begin(); //Now that the unit is reset and I2C inteface mode, we need to begin I2C

	//Configure the si4703 to use the crystal
	readRegisters(); //Read the current register set
	si4703_registers[0x07] = 0x8100; //Enable the oscillator, from AN230 page 9, rev 0.61 (works)
	updateRegisters(); //Update
	delay(500); //Wait for clock to settle - from AN230 page 9

	//Power up the si4703 with our default settings
	//(No Mute, RDS Enabled, 200kHz channel spacing, lowest volume
	readRegisters(); //Read the current register set
	si4703_registers[POWERCFG] = 0x4001; //Enable the IC
	si4703_registers[SYSCONFIG1] |= (1<<RDS); //Enable RDS

#ifdef IN_EUROPE
	si4703_registers[SYSCONFIG1] |= (1<<DE); //50kHz Europe setup
	si4703_registers[SYSCONFIG2] |= (1<<SPACE0); //100kHz channel spacing for Europe
#else
	si4703_registers[SYSCONFIG2] &= ~(1<<SPACE1 | 1<<SPACE0) ; //Force 200kHz channel spacing for USA
#endif

	si4703_registers[SYSCONFIG2] &= 0xFFF0; //Clear volume bits
	si4703_registers[SYSCONFIG2] |= 0x0001; //Set volume to lowest
	updateRegisters(); //Update

	//Give the si4703 enough time to power up.
	delay(110); //Max powerup time, from datasheet page 13
	
}

/*
* Description: 
*	Used to to tune the radio to a desired frequency. The library uses the mode indicated in the
* 	begin() function to determine how to set the frequency.
* Parameters:
*	frequency - The frequency to tune to, in kHz (or in 10kHz if using FM mode).
* Returns:
*	True
* TODO:
* 	Make the function return true if the tune was successful, else return false.
*/
void Si4703::tuneFrequency(int newChannel){
	//Freq(MHz) = 0.200(in USA) * Channel + 87.5MHz
	//97.3 = 0.2 * Chan + 87.5
	//9.8 / 0.2 = 49
	newChannel *= 10; //973 * 10 = 9730
	newChannel -= 8750; //9730 - 8750 = 980

#ifdef IN_EUROPE
	newChannel /= 10; //980 / 10 = 98
#else
	newChannel /= 20; //980 / 20 = 49
#endif

	//These steps come from AN230 page 20 rev 0.5
	readRegisters();
	si4703_registers[CHANNEL] &= 0xFE00; //Clear out the channel bits
	si4703_registers[CHANNEL] |= newChannel; //Mask in the new channel
	si4703_registers[CHANNEL] |= (1<<TUNE); //Set the TUNE bit to start
	updateRegisters();

	//delay(60); //Wait 60ms - you can use or skip this delay

	//Poll to see if STC is set
	while(1) {
		readRegisters();
		if( (si4703_registers[STATUSRSSI] & (1<<STC)) != 0) break; //Tuning complete!
		Serial.println("Tuning");
	}

	readRegisters();
	si4703_registers[CHANNEL] &= ~(1<<TUNE); //Clear the tune after a tune has completed
	updateRegisters();

	//Wait for the si4703 to clear the STC as well
	while(1) {
		readRegisters();
		if( (si4703_registers[STATUSRSSI] & (1<<STC)) == 0) break; //Tuning complete!
		Serial.println("Waiting...");
	}
}
/*
* Description:
*	This function currently does not work!
* TODO:
*	Make this function work.
*/
int Si4703::getFrequency(void){
	readRegisters();
	int channel = si4703_registers[READCHAN] & 0x03FF; //Mask out everything but the lower 10 bits

#ifdef IN_EUROPE
	//Freq(MHz) = 0.100(in Europe) * Channel + 87.5MHz
	//X = 0.1 * Chan + 87.5
	channel *= 1; //98 * 1 = 98 - I know this line is silly, but it makes the code look uniform
#else
	//Freq(MHz) = 0.200(in USA) * Channel + 87.5MHz
	//X = 0.2 * Chan + 87.5
	channel *= 2; //49 * 2 = 98
#endif

	channel += 875; //98 + 875 = 973
	return(channel);
}
/*
* Description:
*	Seeks up or down to the next valid channel.
* Parameters:
*	direction - The direction that the radio will search for a new channel (UP or DOWN are valid parameters)
* Returns:
*	True is a valid channel was found, otherwise false
*/
bool Si4703::seek(bool seekDirection){
	readRegisters();

	//Set seek mode wrap bit
	//si4703_registers[POWERCFG] |= (1<<SKMODE); //Allow wrap
	si4703_registers[POWERCFG] &= ~(1<<SKMODE); //Disallow wrap - if you disallow wrap, you may want to tune to 87.5 first

	if(seekDirection == DOWN) si4703_registers[POWERCFG] &= ~(1<<SEEKUP); //Seek down is the default upon reset
	else si4703_registers[POWERCFG] |= 1<<SEEKUP; //Set the bit to seek up

	si4703_registers[POWERCFG] |= (1<<SEEK); //Start seek

	updateRegisters(); //Seeking will now start

	//Poll to see if STC is set
	while(1) {
		readRegisters();
		if((si4703_registers[STATUSRSSI] & (1<<STC)) != 0) break; //Tuning complete!

		Serial.print("Trying station:");
		Serial.println(getFrequency());
	}

	readRegisters();
	int valueSFBL = si4703_registers[STATUSRSSI] & (1<<SFBL); //Store the value of SFBL
	si4703_registers[POWERCFG] &= ~(1<<SEEK); //Clear the seek bit after seek has completed
	updateRegisters();

	//Wait for the si4703 to clear the STC as well
	while(1) {
		readRegisters();
		if( (si4703_registers[STATUSRSSI] & (1<<STC)) == 0) break; //Tuning complete!
		Serial.println("Waiting...");
	}

	if(valueSFBL) { //The bit was set indicating we hit a band limit or failed to find a station
		Serial.println("Seek limit hit"); //Hit limit of band during seek
		return(FAIL);
	}

	Serial.println("Seek complete"); //Tuning complete!
	return(SUCCESS);
}
/*
* Description:
*	Increases or decreases the current volume
* Parameters:
*	setting - Determines is volume will be increased or decreased (UP or DOWN are valid parameters)
*/
void Si4703::volume(bool setting){
	byte currentVolume=0;
	
	//Now update the volume setting in the configuration register
	readRegisters(); //Read the current register set
	currentVolume = si4703_registers[SYSCONFIG2] & 0x000F; //Read the current volume level
	if((setting == UP) && (currentVolume < 16))currentVolume+=1;
	else if((setting == DOWN) && (currentVolume > 0))currentVolume -=1;
	si4703_registers[SYSCONFIG2] &= 0xFFF0; //Clear volume bits
	si4703_registers[SYSCONFIG2] |= currentVolume; //Set new volume
	updateRegisters(); //Update
}
/*
* Description:
*	Toggles the mute setting
*/
void Si4703::toggleMute(void){
	readRegisters();
	si4703_registers[POWERCFG] ^= (1<<DMUTE); //Toggle Mute bit
	updateRegisters();
}


/****************************************************************
*
* Private Functions
*
****************************************************************/
/*
* Send the register values in the local copy to the si4703
*/
bool Si4703::updateRegisters(void){
	Wire.beginTransmission(SI4703);
	//A write command automatically begins with register 0x02 so no need to send a write-to address
	//First we send the 0x02 to 0x07 control registers
	//In general, we should not write to registers 0x08 and 0x09
	for(int regSpot = 0x02 ; regSpot < 0x08 ; regSpot++) {
		byte high_byte = si4703_registers[regSpot] >> 8;
		byte low_byte = si4703_registers[regSpot] & 0x00FF;

		Wire.send(high_byte); //Upper 8 bits
		Wire.send(low_byte); //Lower 8 bits
	}

	//End this transmission
	byte ack = Wire.endTransmission();
	if(ack != 0) { //We have a problem! 
		Serial.print("Write Fail:"); //No ACK!
		Serial.println(ack, DEC); //I2C error: 0 = success, 1 = data too long, 2 = rx NACK on address, 3 = rx NACK on data, 4 = other error
		return(FAIL);
	}

	return(SUCCESS);
}
/*
* Copy the register values in the si4703 to the local array
*/
void Si4703::readRegisters(void){
	//Si4703 begins reading from register upper register of 0x0A and reads to 0x0F, then loops to 0x00.
	Wire.requestFrom(SI4703, 32); //We want to read the entire register set from 0x0A to 0x09 = 32 bytes.

	while(Wire.available() < 32) ; //Wait for 16 words/32 bytes to come back from slave I2C device
	//We may want some time-out error here

	//Remember, register 0x0A comes in first so we have to shuffle the array around a bit
	for(int x = 0x0A ; ; x++) { //Read in these 32 bytes
		if(x == 0x10) x = 0; //Loop back to zero
		si4703_registers[x] = Wire.receive() << 8;
		si4703_registers[x] |= Wire.receive();
		if(x == 0x09) break; //We're done!
	}
}